/*++
Copyright (c) Microsoft Corporation.  All rights reserved.

    THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF ANY
    KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
    IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR
    PURPOSE.

Module Name:

    install.c

Abstract:

    Win32 routines to dynamically load and unload a Windows NT kernel-mode
    driver using the Service Control Manager APIs.

Environment:

    User mode only

--*/


#include <windows.h>
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strsafe.h>
#include "public.h"


#define SYSTEM32_DRIVERS "\\System32\\Drivers\\"


BOOLEAN InstallDriver(__in SC_HANDLE SchSCManager, __in LPCTSTR DriverName, __in LPCTSTR ServiceExe);
BOOLEAN RemoveDriver(__in SC_HANDLE SchSCManager, __in LPCTSTR DriverName);
BOOLEAN StartDriver(__in SC_HANDLE SchSCManager, __in LPCTSTR DriverName);
BOOLEAN StopDriver(__in SC_HANDLE SchSCManager, __in LPCTSTR DriverName);
BOOLEAN OpenCloseService(__in SC_HANDLE SchSCManager, __in LPCTSTR DriverName);


int open_cntr, close_cntr;

/* ------------------------------------------------------------------------- */
void wait4key(void)
   {
   while (!_kbhit())
      {
      putchar('.');
      Sleep(1000);
      }
   _getch();
   printf("\n");
   }

/* ------------------------------------------------------------------------- */
BOOLEAN 
InstallDriver(__in SC_HANDLE SchSCManager, __in LPCTSTR DriverName, __in LPCTSTR ServiceExe)
   {
   SC_HANDLE   schService;
   DWORD       err;

   printf("---InstallDriver\n");
   //
   // NOTE: This creates an entry for a standalone driver. If this
   //       is modified for use with a driver that requires a Tag,
   //       Group, and/or Dependencies, it may be necessary to
   //       query the registry for existing driver information
   //       (in order to determine a unique Tag, etc.).
   //

   //
   // Create a new a service object.
   //

   schService = CreateService(SchSCManager,           // handle of service control manager database
                              DriverName,             // address of name of service to start
                              DriverName,             // address of display name
                              SERVICE_ALL_ACCESS,     // type of access to service
                              SERVICE_KERNEL_DRIVER,  // type of service
                              SERVICE_AUTO_START,     // when to start service
//                            SERVICE_DEMAND_START,   // when to start service
                              SERVICE_ERROR_NORMAL,   // severity if service fails to start
                              ServiceExe,             // address of name of binary file
                              NULL,                   // service does not belong to a group
                              NULL,                   // no tag requested
                              NULL,                   // no dependency names
                              NULL,                   // use LocalSystem account
                              NULL                    // no password for service account
                              );

   if (schService == NULL) 
      {
      err = GetLastError();
      if (err == ERROR_SERVICE_EXISTS) 
         {
         // Ignore this error.
         printf("CreateService: Service Exists!\n");
         return TRUE;
         } 
      
      if (err == ERROR_SERVICE_MARKED_FOR_DELETE) 
         // Previous instance of the service is not fully deleted so sleep and try again.
         printf("Previous instance of the service is not fully deleted. Try again...\n");
      else 
         // Indicate an error.
         printf("CreateService failed!  Error = %d %s\n", err, strerror(err));
      return FALSE;
      }

   open_cntr++;
      
   // Close the service object.
   CloseServiceHandle(schService);
   close_cntr++;

   // Indicate success.
   return TRUE;
   }   // InstallDriver

   
/* ------------------------------------------------------------------------- */
BOOLEAN 
ManageDriver(__in LPCTSTR DriverName, __in LPCTSTR ServiceName, __in USHORT Function)
   {

   SC_HANDLE  schSCManager;
   BOOLEAN    rCode = TRUE;
   DWORD      err;

   // Insure (somewhat) that the driver and service names are valid.
   if (!DriverName || !ServiceName)
      {
      printf("Invalid Driver or Service provided to ManageDriver() \n");
      return FALSE;
      }

   // Connect to the Service Control Manager and open the Services database.
   schSCManager = OpenSCManager(
                               NULL,                   // local machine
                               NULL,                   // local database
                               SC_MANAGER_ALL_ACCESS   // access required
                               );
   if (!schSCManager)
      {
      err = GetLastError();
      printf("Open SC Manager failed! Error = %d %s\n", err, strerror(err));
      return FALSE;
      }

   open_cntr++;
   // Do the requested function.
   switch( Function )
      {
      case DRIVER_FUNC_INSTALL:
         if (rCode = InstallDriver(schSCManager, DriverName, ServiceName)) // Install the driver service.
            rCode = StartDriver(schSCManager, DriverName);                 // Start the driver service (i.e. start the driver).
         break;

      case DRIVER_FUNC_REMOVE:
         if (rCode = StopDriver(schSCManager, DriverName))           // Stop the driver.
            rCode = RemoveDriver(schSCManager, DriverName);         // Remove the driver service.
         break;

      default:
         printf("Unknown ManageDriver() function. \n");
         rCode = FALSE;
         break;
      }

   // Close handle to service control manager.
   if (schSCManager)
      {
      CloseServiceHandle(schSCManager);
      close_cntr++;
      }
   return rCode;
   }   // ManageDriver


/* ------------------------------------------------------------------------- */
BOOLEAN
RemoveDriver(__in SC_HANDLE SchSCManager, __in LPCTSTR DriverName)
   {
   SC_HANDLE   schService;
   BOOLEAN     rCode;
   DWORD       err;

   printf("---RemoveDriver\n");
   
   // Open the handle to the existing service.
   schService = OpenService(SchSCManager, DriverName, SERVICE_ALL_ACCESS);
   if (schService == NULL)
      {
      err = GetLastError();
      printf("OpenService failed!  Error = %d %s\n", err, strerror(err));
      return FALSE;
      }

//   open_cntr++;
   
   // Mark the service for deletion from the service control manager database.
   if (DeleteService(schService))
      {
      // Indicate success.
      rCode = TRUE;
//      printf("DeleteService successful\n");
      } 
   else
      {
      err = GetLastError();
      printf("DeleteService failed!  Error = %d %s\n", err, strerror(err));
      // Indicate failure.  Fall through to properly close the service handle.
      rCode = FALSE;
      }

   // Close the service object.
   if (!CloseServiceHandle(schService))
      {
      err = GetLastError();
      printf("CloseServiceHandle!  Error = %d %s\n", err, strerror(err));
      rCode = FALSE;
      }
   else
      {
//      printf("CloseServiceHandle called successfully\n");
//      close_cntr++;
      }

//   printf("Exiting RemoveDriver..\n\n");
   
   return rCode;
   }   // RemoveDriver


/* ------------------------------------------------------------------------- */
BOOLEAN
StartDriver(__in SC_HANDLE SchSCManager, __in LPCTSTR DriverName)
   {
   SC_HANDLE   schService;
   DWORD       err;

   printf("---StartDriver\n");
   
   // Open the handle to the existing service.
   schService = OpenService(SchSCManager, DriverName, SERVICE_ALL_ACCESS);
   if (schService == NULL) 
      {
      err = GetLastError();
      printf("OpenService failed!  Error = %d %s\n", err, strerror(err));
      return FALSE;
      }

//   open_cntr++;
   // Start the execution of the service (i.e. start the driver).
   if (!StartService(
                    schService,     // service identifier
                    0,              // number of arguments
                    NULL            // pointer to arguments
                    ))
      {
      err = GetLastError();
      if (err == ERROR_SERVICE_ALREADY_RUNNING)
         {
         // Ignore this error.
         return TRUE;
         } 
      else 
         {
         printf("StartService failure! Error = %d %s\n", err, strerror(err));
         return FALSE;
         }
      }

   // Close the service object.
   CloseServiceHandle(schService);
//   close_cntr++;

//   printf("\n");
   return TRUE;
   }   // StartDriver


/* ------------------------------------------------------------------------- */
BOOLEAN
StopDriver(__in SC_HANDLE SchSCManager, __in LPCTSTR DriverName)
   {
   BOOLEAN         rCode = TRUE;
   SC_HANDLE       schService;
   SERVICE_STATUS  serviceStatus;
   DWORD           err;

   printf("---StopDriver\n");
   
   // Open the handle to the existing service.
   schService = OpenService(SchSCManager, DriverName, SERVICE_ALL_ACCESS);
   if (schService == NULL) 
      {
      err = GetLastError();
      printf("OpenService failed!  Error = %d %s\n", err, strerror(err));
      return FALSE;
      }

   open_cntr++;

   // Request that the service stop.
   if (ControlService(schService, SERVICE_CONTROL_STOP, &serviceStatus)) 
      {
      rCode = TRUE;
//      printf("ControlService successfully stopped services!\n");
      } 
   else 
      {
      err = GetLastError();
      printf("ControlService failed to stop services!");
      rCode = FALSE;
      }

   // Close the service object.
   if (CloseServiceHandle(schService))
      {
//      printf("CloseServiceHandle called successfully\n");
//      close_cntr++;
      }
   else
      {
      err = GetLastError();
      printf("CloseServiceHandle!  Error = %d %s\n", err, strerror(err));
      rCode = FALSE;
      }
//   printf("Exiting StopDriver..\n\n");
//   printf("\nopen_cntr =%d close_cntr =%d\n", open_cntr, close_cntr);
   return rCode;
   }  //  StopDriver


/* ------------------------------------------------------------------------- */
// Caller must free returned pathname string.
PCHAR
BuildDriversDirPath(__in PSTR DriverName)
   {
   size_t  remain;
   size_t  len;
   PCHAR   dir;

   if (!DriverName || strlen(DriverName) == 0)
      return NULL;

   remain = MAX_PATH;
   // Allocate string space
   dir = (PCHAR) malloc( remain + 1 );
   if (!dir)
      return NULL;

   while (1)
      {
      // Get the base windows directory path.
      len = GetWindowsDirectory( dir, (UINT) remain );
      if (len == 0 || (remain - len) < sizeof(SYSTEM32_DRIVERS)) 
         break;

      remain -= len;
      // Build dir to have "%windir%\System32\Drivers\<DriverName>".
      if (FAILED( StringCchCat(dir, remain, SYSTEM32_DRIVERS) )) 
         break;

      remain -= sizeof(SYSTEM32_DRIVERS);
      len    += sizeof(SYSTEM32_DRIVERS);
      len    += strlen(DriverName);

      if (remain < len)
         break;

      if (FAILED( StringCchCat(dir, remain, DriverName))) 
         break;

      dir[len] = '\0';  // keeps prefast happy
      return dir;
      }

   free(dir);
   return NULL;
   }

/* ------------------------------------------------------------------------- */
BOOLEAN
SetupDriverName(__inout_bcount_full(BufferLength) PCHAR DriverLocation, __in ULONG BufferLength)
   {
   HANDLE fileHandle;
   DWORD  driverLocLen = 0;
   BOOL   ok;
   PCHAR  driversDir;
   DWORD  err;

   // Get the current directory.
   printf("---SetupDriverName\n");
   
   driverLocLen = GetCurrentDirectory(BufferLength, DriverLocation);
   if (driverLocLen == 0) 
      {
      err = GetLastError();
      printf("GetCurrentDirectory failed!  Error = %d %s\n", err, strerror(err));
      return FALSE;
      }

//   printf("Current directory:  %s\n", DriverLocation);
   // Setup path name to driver file.
   if (FAILED( StringCbCat(DriverLocation, BufferLength, "\\"DRIVER_NAME".sys") )) 
       {
       printf("StringCbCat failed:  %s\n", DriverLocation);
       return FALSE;
       }

//   printf("StringCbCat:  %s\n", DriverLocation);

   // Insure driver file is in the specified directory.
   fileHandle  = CreateFile(
                           DriverLocation,
                           GENERIC_READ,
                           FILE_SHARE_READ,
                           NULL,
                           OPEN_EXISTING,
                           FILE_ATTRIBUTE_NORMAL,
                           NULL 
                           );

   if (fileHandle == INVALID_HANDLE_VALUE) 
      {
      printf("Driver: %s.SYS is not in the %s directory.\n", DRIVER_NAME, DriverLocation);
      return FALSE;
      }

   // Build %windir%\System32\Drivers\<DRIVER_NAME> path.
   driversDir = BuildDriversDirPath(DRIVER_NAME ".sys");
   if (!driversDir) 
      {
      printf("BuildDriversDirPath failed!\n");
      CloseHandle(fileHandle);
      return FALSE;
      }

//   printf("BuildDriversDirPath:  %s\n", driversDir);

   // Copy the driver to %windir%\system32\drivers
   ok = CopyFile(DriverLocation, driversDir, FALSE);
   if (!ok) 
      {
      err = GetLastError();
      printf("CopyFile failed:  error %d %s - driversDir %s\n", err, strerror(err), driversDir);
      free(driversDir);
      CloseHandle(fileHandle);
      return FALSE;
      }

   if (FAILED(StringCchCopy(DriverLocation, BufferLength, driversDir))) 
      {
      free(driversDir);
      CloseHandle(fileHandle);
      return FALSE;
      }

//   printf("StringCchCopy:  %s\n", DriverLocation);

   free(driversDir);

   // Close open file handle.
   CloseHandle(fileHandle);

   // Indicate success.
   return TRUE;
   }  // SetupDriverName


   
/* ------------------------------------------------------------------------- */
/* ------------------------------------------------------------------------- */
VOID __cdecl
main(__in ULONG argc, __in_ecount(argc) PCHAR argv[])
   {
   TCHAR  driverLocation[MAX_PATH];
   HANDLE hDevice;
   DWORD  err = 0;
   ULONG  ulReturnedLength;
   PCHAR  pPar;
   PCHAR  driversDir = NULL;
   
   if (argc != 2)
      {
      printf("Need one argument:  \"inst\" for install or \"rem\" for remove\n");
      exit(1);
      }
   
   pPar = argv[1];
   if (strcmp(pPar, "inst") && strcmp(pPar, "rem"))
      {
      printf("unknown argument %s\n", pPar);
      exit(1);
      }

   // try to open the device
   hDevice =  CreateFile(
                        "\\\\.\\XGpioDev",
                        GENERIC_READ | GENERIC_WRITE,
                        0,
                        NULL,
                        CREATE_ALWAYS,
                        FILE_ATTRIBUTE_NORMAL,
                        NULL
                        );

   if (hDevice == INVALID_HANDLE_VALUE)
      {
//      printf("hDevice is INVALID_HANDLE_VALUE\n");
      err = GetLastError();
      if (!(err == ERROR_FILE_NOT_FOUND || err == ERROR_PATH_NOT_FOUND))
         {
         printf("CreateFile failed!  ERROR_FILE_NOT_FOUND = %d %s\n", err, strerror(err));
         return ;
         }

      // The driver is not started yet so let us the install the driver.
      while (1)
         {
         if (!strcmp(pPar, "rem"))
            {
            printf("Driver not installed\n");
            break;
            }
         
         // First setup full path to driver name.
         if (!SetupDriverName(driverLocation, sizeof(driverLocation)))
            {
            printf("driver name not found in the current subdirectory\n");
            break; 
            }

         if (ManageDriver(DRIVER_NAME, driverLocation, DRIVER_FUNC_INSTALL) == 0) 
            {
            printf("Unable to install driver. \n");
            // Error - remove driver.
            ManageDriver(DRIVER_NAME, driverLocation, DRIVER_FUNC_REMOVE);
            break;
            }

         // open the device
         hDevice =  CreateFile(
                              "\\\\.\\XGpioDev",
                              GENERIC_READ | GENERIC_WRITE,
                              0,
                              NULL,
                              CREATE_ALWAYS,
                              FILE_ATTRIBUTE_NORMAL,
                              NULL
                              );
         if (hDevice == INVALID_HANDLE_VALUE) 
            {
            err = GetLastError();
            printf ( "Error: CreatFile Failed : %d %s\n", err, strerror(err));
            break;
            }
   
         printf("Driver installed\n");
         CloseHandle(hDevice);
         break;
         }
      return;
      }


//   printf("hDevice is VALID handle.\n");
   while(1)
      {
      if (!strcmp(pPar, "inst"))
         {
         printf("Driver already installed\n");
         break;
         }
      
      driversDir = BuildDriversDirPath(DRIVER_NAME ".sys");
      if (driversDir == 0)
         {
         printf("BuildDriversDirPath failed!\n");
         break;
         }

      if (FAILED(StringCchCopy(driverLocation, sizeof(driverLocation), driversDir))) 
         break;

 //     printf("driverLocation:   %s\n", driverLocation);
      if (ManageDriver(DRIVER_NAME, driverLocation, DRIVER_FUNC_REMOVE) == FALSE)
         break;
         
      if (!DeleteFile(driverLocation))
         {
         err = GetLastError();
         printf("DeleteFile failed:  error %d %s - driverLocation %s\n", err, strerror(err), driverLocation);
         break;
         }

      printf("Driver deleted\n\n");
      printf("Reboot the system to complete driver's removal\n");
      break;
      }
 
   if (driversDir)
      free(driversDir);

   if (hDevice)
      CloseHandle(hDevice);
      

//   printf("\nopen_cntr =%d close_cntr =%d\n", open_cntr, close_cntr);
   }
