/*++
// xgpioport.c - P60xBO GPIO driver source file
//
// Copyright (c) 2007 General Micro Systems, Inc.  All Rights Reserved
//
// You are hereby granted a copyright license to use, modify and
// distribute this SOFTWARE so long as the entire notice is retained
// without alteration in any modified and/or redistributed versions,
// and that such modified versions are clearly identified as such.
// No licenses are granted by implication, estopple or otherwise under
// any patents or trademarks of General Micro Systems, Inc (GMS).
//
// The SOFTWARE is provided on an "AS IS" basis and without warrenty,
// to the maximum extent permitted by applicable law.
//
// GMS DISCLAIMS ALL WARRANTIES WHETHER EXPRESS OR IMPLIED, INCLUDING
// WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE
// AND ANY WARRANTY AGAINST INFRINGEMENT WITH REGARD TO THE SOFTWARE
// (INCLUDING ANY MODIFIED VERSIONS THEREOF) AND ANY ACCOMPANYING
// WRITTEN MATERIAL.
//
// To the maximum extent permitted by applicable law, IN NO EVENT SHALL
// GMS BE LIABLE FOR ANY DAMAGE WHATSOEVER (INCLUDING WITHOUT LIMITATION,
// DAMAGES FOR LOSS OF BUSINESS PROFITS, BUSINESS INTERRUPTION, LOSS OF
// BUSINESS INFORMATION, OR OTHER PECUNIARY LOSS) ARISING FROM THE USE
// OR INABILITY TO USE THE SOFTWARE.  GMS assumes no resposibility for
// for the maintenance or support of the SOFTWARE
//

// Maintenance Log
//---------------------------------------------------------------------
// v1.7     31-Oct-2013  vp     initial non .inf windows driver
// --------------------------------------------------------------------
--*/


//
// Include files.
//

#include <ntddk.h>          // various NT definitions
#include <string.h>

#include "gpioctl.h"



#define GPIO_DEVICE_NAME L"\\Device\\XGpio"
#define DOS_DEVICE_NAME  L"\\DosDevices\\XGpioDev"


#define DBG    1
#if DBG
#define DebugPrint(_x_) \
               DbgPrint ("PortIo:"); \
               DbgPrint _x_;

#else
#define DebugPrint(_x_)
#endif



DRIVER_INITIALIZE DriverEntry;

__drv_dispatchType(IRP_MJ_CREATE)
__drv_dispatchType(IRP_MJ_CLOSE)
DRIVER_DISPATCH GpioCreateClose;

__drv_dispatchType(IRP_MJ_DEVICE_CONTROL)
DRIVER_DISPATCH GpioDeviceControl;

DRIVER_UNLOAD   GpioUnloadDriver;


VOID
PrintIrpInfo(
    PIRP Irp
    );

VOID
PrintChars(
    __in_ecount(CountChars) PCHAR BufferAddress,
    __in size_t CountChars
    );

#ifdef ALLOC_PRAGMA
#pragma alloc_text( INIT, DriverEntry )
#pragma alloc_text( PAGE, GpioCreateClose)
#pragma alloc_text( PAGE, GpioUnloadDriver)
#pragma alloc_text( PAGE, GpioDeviceControl)
#pragma alloc_text( PAGE, PrintIrpInfo)
#pragma alloc_text( PAGE, PrintChars)
#endif // ALLOC_PRAGMA


/* -------------------------------------------------------------------------- */
NTSTATUS
DriverEntry(
    __in PDRIVER_OBJECT   DriverObject,
    __in PUNICODE_STRING  RegistryPath
    )
/*++

Routine Description:
    This routine is called by the Operating System to initialize the driver.
    It creates the device object, fills in the dispatch entry points and
    completes the initialization.

Arguments:
    DriverObject - a pointer to the object that represents this device
    driver.
    RegistryPath - a pointer to our Services key in the registry.

Return Value:
    STATUS_SUCCESS if initialized; an error otherwise.

--*/

{
    NTSTATUS        ntStatus;
    UNICODE_STRING  ntUnicodeString;        // GPIO Device Name "\Device\XGpio"
    UNICODE_STRING  ntWin32NameString;      // Win32 Name "\DosDevices\XGpioDev"
    PDEVICE_OBJECT  deviceObject = NULL;    // ptr to device object

    UNREFERENCED_PARAMETER(RegistryPath);

    DebugPrint(("Entered Driver Entry\n"));
    
    RtlInitUnicodeString( &ntUnicodeString, GPIO_DEVICE_NAME );

    ntStatus = IoCreateDevice(
        DriverObject,                  // Our Driver Object
        0,                             // We don't use a device extension
        &ntUnicodeString,              // Device name "\Device\Xgpio"
        FILE_DEVICE_UNKNOWN,                     // Device type
        FILE_DEVICE_SECURE_OPEN,       // Device characteristics
        FALSE,                         // Not an exclusive device
        &deviceObject );               // Returned ptr to Device Object

    if ( !NT_SUCCESS( ntStatus ) )
    {
        DebugPrint(("Couldn't create the device object\n"));
        return ntStatus;
    }

    // Initialize the driver object with this driver's entry points.
    //
    DriverObject->MajorFunction[IRP_MJ_CREATE]         = GpioCreateClose;
    DriverObject->MajorFunction[IRP_MJ_CLOSE]          = GpioCreateClose;
    DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = GpioDeviceControl;
    DriverObject->DriverUnload                         = GpioUnloadDriver;

    // Initialize a Unicode String containing the Win32 name for our device.
    //
    RtlInitUnicodeString( &ntWin32NameString, DOS_DEVICE_NAME );

    // Create a symbolic link between our device name  and the Win32 name
    //
    ntStatus = IoCreateSymbolicLink(
                        &ntWin32NameString, &ntUnicodeString );

    if ( !NT_SUCCESS( ntStatus ) )
    {
        // Delete everything that this routine has allocated.
        //
        DebugPrint(("Couldn't create symbolic link\n"));
        IoDeleteDevice( deviceObject );
    }

    return ntStatus;
}


/* -------------------------------------------------------------------------- */
NTSTATUS
GpioCreateClose(
    PDEVICE_OBJECT DeviceObject,
    PIRP Irp
    )
/*++

Routine Description:
    This routine is called by the I/O system when the SIOCTL is opened or
    closed.
    No action is performed other than completing the request successfully.

Arguments:
    DeviceObject - a pointer to the object that represents the device
    that I/O is to be done on.
    Irp - a pointer to the I/O Request Packet for this request.

Return Value:
    NT status code

--*/

{
    UNREFERENCED_PARAMETER(DeviceObject);

    PAGED_CODE();

    Irp->IoStatus.Status = STATUS_SUCCESS;
    Irp->IoStatus.Information = 0;

    IoCompleteRequest( Irp, IO_NO_INCREMENT );

    return STATUS_SUCCESS;
}


/* -------------------------------------------------------------------------- */
NTSTATUS
GpioDeviceControl(
    PDEVICE_OBJECT DeviceObject,
    PIRP Irp
    )
/*++

Routine Description:
    This routine is called by the I/O system to perform a device I/O
    control function.

Arguments:

    DeviceObject - a pointer to the object that represents the device
    that I/O is to be done on.
    Irp - a pointer to the I/O Request Packet for this request.

Return Value:
    NT status code

--*/

{
    PIO_STACK_LOCATION  pIrpStack;
    NTSTATUS            Status;
    PUCHAR  pIOBuffer;            // Pointer to transfer buffer
    ULONG   InBufferSize;         // Amount of data avail. from caller.
    ULONG   OutBufferSize;        // Max data that caller can accept.
    ULONG   DataBufferSize;
    PVOID   portBase;
    UCHAR   regVal;
    
    UNREFERENCED_PARAMETER(DeviceObject);

    PAGED_CODE();
//    DebugPrint(("Entered GpioDispatch\n"));

    pIrpStack = IoGetCurrentIrpStackLocation(Irp);

    Status = STATUS_SUCCESS;

    portBase       = 0x0A00;
    pIOBuffer      = (PUCHAR) Irp->AssociatedIrp.SystemBuffer;
    DataBufferSize = sizeof(XGPIOPORT_WRITE_INPUT);
    Irp->IoStatus.Information = 0;

    switch (pIrpStack->Parameters.DeviceIoControl.IoControlCode)
    {
        case IOCTL_P60xBO_XGPIO_GET_DIR:
             portBase = 0x0A01;
        case IOCTL_P60xBO_XGPIO_GET_LEVEL:
            OutBufferSize = pIrpStack->Parameters.DeviceIoControl.OutputBufferLength;
            if ( OutBufferSize == 0 || OutBufferSize < DataBufferSize )
            {
                Status = STATUS_INVALID_PARAMETER;
                goto End;
            }
            regVal = READ_PORT_UCHAR( (PUCHAR)((ULONG_PTR)portBase) );

           *(PUCHAR)pIOBuffer = regVal;
            Irp->IoStatus.Information = DataBufferSize;

            DebugPrint (("GpioReadPort port =0x%08x read =0x%02x\n", (ULONG_PTR)portBase, *(PUCHAR)pIOBuffer));
            break;

        case IOCTL_P60xBO_XGPIO_SET_DIR:
             portBase = 0x0A01;
        case IOCTL_P60xBO_XGPIO_SET_LEVEL:
            InBufferSize = pIrpStack->Parameters.DeviceIoControl.InputBufferLength;
            if ( InBufferSize == 0 || InBufferSize <  DataBufferSize )
            {
                Status = STATUS_INVALID_PARAMETER;
                goto End;
            }
            regVal = READ_PORT_UCHAR( (PUCHAR)((ULONG_PTR)portBase) );
            DebugPrint (("GpioIoctlWritePort: port =0x%04x read  =0x%02x\n", (ULONG_PTR)portBase, regVal));
            regVal &= ~pIOBuffer[1] ;
            regVal |= pIOBuffer[0] & pIOBuffer[1];
            DebugPrint (("GpioIoctlWritePort: port =0x%04x write =0x%02x\n", (ULONG_PTR)portBase, regVal));
            WRITE_PORT_UCHAR( (PUCHAR)((ULONG_PTR)portBase), regVal );
            break;

        default:
            Status = STATUS_INVALID_DEVICE_REQUEST;
            DebugPrint(("ERROR: unrecognized IOCTL %x\n", pIrpStack->Parameters.DeviceIoControl.IoControlCode));
            break;

    }

End:

    // Finish the I/O operation by simply completing the packet and returning
    // the same status as in the packet itself.
    Irp->IoStatus.Status = Status;

    IoCompleteRequest( Irp, IO_NO_INCREMENT );

    return Status;
}


/* -------------------------------------------------------------------------- */
VOID
GpioUnloadDriver(
    __in PDRIVER_OBJECT DriverObject
    )
/*++

Routine Description:
    Free all the allocated resources, etc.

Arguments:
    DriverObject - pointer to a driver object.

Return Value:
    VOID.

--*/
{

    PDEVICE_OBJECT deviceObject = DriverObject->DeviceObject;
    UNICODE_STRING uniWin32NameString;

    PAGED_CODE();

    DebugPrint(("Entered GpioUnload\n"));
    // Create counted string version of our Win32 device name.
    //
    RtlInitUnicodeString( &uniWin32NameString, DOS_DEVICE_NAME );
    DebugPrint(("After RtlInitUnicodeString:  %s\n", uniWin32NameString));

    // Delete the link from our device name to a name in the Win32 namespace.
    //
    IoDeleteSymbolicLink( &uniWin32NameString );
    if ( deviceObject != NULL )
    {
        IoDeleteDevice( deviceObject );
        DebugPrint(("IoDeleteDevice called\n"));
    }
    else
    {
        DebugPrint(("IoDeleteDevice not called, NULL deviceObject \n"));
    }
}


/* -------------------------------------------------------------------------- */
VOID
PrintIrpInfo(
    PIRP Irp)
{
    PIO_STACK_LOCATION  irpSp;
    irpSp = IoGetCurrentIrpStackLocation( Irp );

    PAGED_CODE();

    DebugPrint(("\tIrp->AssociatedIrp.SystemBuffer = 0x%p\n",
        Irp->AssociatedIrp.SystemBuffer));
    DebugPrint(("\tIrp->UserBuffer = 0x%p\n", Irp->UserBuffer));
    DebugPrint(("\tirpSp->Parameters.DeviceIoControl.Type3InputBuffer = 0x%p\n",
        irpSp->Parameters.DeviceIoControl.Type3InputBuffer));
    DebugPrint(("\tirpSp->Parameters.DeviceIoControl.InputBufferLength = %d\n",
        irpSp->Parameters.DeviceIoControl.InputBufferLength));
    DebugPrint(("\tirpSp->Parameters.DeviceIoControl.OutputBufferLength = %d\n",
        irpSp->Parameters.DeviceIoControl.OutputBufferLength ));
    return;
}


/* -------------------------------------------------------------------------- */
VOID
PrintChars(
    __in_ecount(CountChars) PCHAR BufferAddress,
    __in size_t CountChars
    )
{
    PAGED_CODE();

    if (CountChars) {
        while (CountChars--) {
            if (*BufferAddress > 31 && *BufferAddress != 127) {
                KdPrint (( "%c", *BufferAddress) );
            } else {
                KdPrint(( ".") );
            }
            BufferAddress++;
        }
        KdPrint (("\n"));
    }
    return;
}
